#include "LinuxWindow.hpp"

#include "platform/opengl/OpenGLContext.hpp"

#include "quark/Core.hpp"
#include "quark/Logger.hpp"
#include "quark/events/ApplicationEvent.hpp"
#include "quark/events/KeyEvent.hpp"
#include "quark/events/MouseEvent.hpp"

#include <GLFW/glfw3.h>

namespace qk {
	static bool GLFWInitialized = false;

	static void GLFWErrorCallback(int error, const char* description) {
		QK_CORE_ERROR("GLFW Error ({0}): '{1}'", error, description);
	}

	std::unique_ptr<Window> Window::Create(const WindowConfig& config) {
		return std::make_unique<LinuxWindow>(config);
	}

	LinuxWindow::LinuxWindow(const WindowConfig& config) {
		Init(config);
	}

	LinuxWindow::~LinuxWindow() {
		Destroy();
	}

	void LinuxWindow::Init(const WindowConfig& config) {
		data.title = config.title;
		data.width = config.width;
		data.height = config.height;


		QK_CORE_INFO("Creating window {0} ({1}x{2})", data.title, data.width, data.height);

		if (!GLFWInitialized) {
			// glfwInitHint(GLFW_PLATFORM, GLFW_PLATFORM_X11);
			int success = glfwInit();
			QK_CORE_ASSERT(success, "Could not initialize GLFW");

			glfwSetErrorCallback(GLFWErrorCallback);

			GLFWInitialized = true;
		}
		
		glfwWindowHintString(GLFW_X11_CLASS_NAME, "testclass");

		window = glfwCreateWindow((int)config.width, (int)config.height, config.title.c_str(), nullptr, nullptr);

		context = std::make_unique<OpenGLContext>(window);
		context->Init();

		glfwSetWindowUserPointer(window, &data);

		InstallGLFWCallbacks();

		SetVSync(true);
	}

	void LinuxWindow::Destroy() {
		glfwDestroyWindow(window);
		window = nullptr;
	}

	void LinuxWindow::InstallGLFWCallbacks() {
		glfwSetWindowSizeCallback(window, [](GLFWwindow* window, int width, int height) {
			WindowData* data = reinterpret_cast<WindowData*>(glfwGetWindowUserPointer(window));

			data->width = width;
			data->height = height;

			WindowResizeEvent event(width, height);
			data->eventCallback(event);
		});

		glfwSetWindowCloseCallback(window, [] (GLFWwindow* window) {
			WindowData* data = reinterpret_cast<WindowData*>(glfwGetWindowUserPointer(window));

			WindowCloseEvent event;
			data->eventCallback(event);
		});

		glfwSetKeyCallback(window, [] (GLFWwindow* window, int key, int scancode, int action, int modes) {
			WindowData* data = reinterpret_cast<WindowData*>(glfwGetWindowUserPointer(window));

			switch (action) {
				case GLFW_PRESS: {
					KeyPressedEvent event(key, 0);
					data->eventCallback(event);
				} break;

				case GLFW_RELEASE: {
					KeyReleasedEvent event(key);
					data->eventCallback(event);
				} break;

				case GLFW_REPEAT: {
					KeyPressedEvent event(key, 1);
					data->eventCallback(event);
				} break;
			}
		});

        glfwSetCharCallback(window, [] (GLFWwindow* window, unsigned int key) {
			WindowData* data = reinterpret_cast<WindowData*>(glfwGetWindowUserPointer(window));

            KeyTypedEvent event(key);
            data->eventCallback(event);
        });

		glfwSetMouseButtonCallback(window, [] (GLFWwindow* window, int button, int action, int modes) {
			WindowData* data = reinterpret_cast<WindowData*>(glfwGetWindowUserPointer(window));

			switch (action) {
				case GLFW_PRESS: {
					MouseButtonPressedEvent event(button);
					data->eventCallback(event);
				} break;

				case GLFW_RELEASE: {
					MouseButtonReleasedEvent event(button);
					data->eventCallback(event);
				} break;
			}
		});

		glfwSetCursorPosCallback(window, [] (GLFWwindow* window, double x, double y) {
			WindowData* data = reinterpret_cast<WindowData*>(glfwGetWindowUserPointer(window));

			MouseMovedEvent event((float)x, (float)y);
			data->eventCallback(event);
		});

		glfwSetScrollCallback(window, [] (GLFWwindow* window, double x, double y) {
			WindowData* data = reinterpret_cast<WindowData*>(glfwGetWindowUserPointer(window));

			MouseScrolledEvent event((float)x, (float)y);
			data->eventCallback(event);
		});
	}

	void LinuxWindow::OnUpdate() {
		glfwPollEvents();
		context->SwapBuffers();
	}

	constexpr void LinuxWindow::SetVSync(bool enabled) {
		if (enabled) {
			glfwSwapInterval(1);
		} else {
			glfwSwapInterval(0);
		}

		data.vsync = enabled;
	}

	bool LinuxWindow::HasVSync() const {
		return data.vsync;
	}
}
